#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Files.py - Functions to read and write the input/output files for an archNEMESIS simulation.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability

import os
import textwrap

import numpy as np
import matplotlib.pyplot as plt

import archnemesis as ans
from archnemesis import (
    Atmosphere_0,
    Variables_0,
    Measurement_0,
    Layer_0,
    OptimalEstimation_0,
    CIA_0,
    Scatter_0
)

from archnemesis.Data.planet_data import planet_info

from archnemesis.Models import Models, ModelBase, ModelParameterEntry
from copy import copy

from archnemesis.helpers import h5py_helper, io_helper
from archnemesis.enums import (
    #PlanetEnum, 
    #AtmosphericProfileFormatEnum, 
    #InstrumentLineshape, 
    LayerType,
    WaveUnit, 
    SpectraUnit,
    SpectralCalculationMode, 
    LowerBoundaryCondition, 
    ScatteringCalculationMode, 
    AerosolPhaseFunctionCalculationMode,
    ParaH2Ratio, 
    RayleighScatteringMode,
    LayerIntegrationScheme,
)

import logging
_lgr = logging.getLogger(__name__)

###############################################################################################
###############################################################################################
#                                            GENERIC
###############################################################################################
###############################################################################################

########################################################################################

def file_lines(fname):

    """
    FUNCTION NAME : file_lines()
    
    DESCRIPTION : Returns the number of lines in a given file
    
    INPUTS : 
 
        fname :: Name of the file

    OPTIONAL INPUTS: none
            
    OUTPUTS : 
 
        nlines :: Number of lines in file

    CALLING SEQUENCE:

        nlines = file_lines(fname)

    MODIFICATION HISTORY : Juan Alday (29/04/2019)

    """

    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1


###############################################################################################
###############################################################################################
#                                   archNEMESIS FILES
###############################################################################################
###############################################################################################

# read_input_files_hdf5()
# read_retparam_hdf5()
# read_bestfit_hdf5()

###############################################################################################

def read_input_files_hdf5(runname,calc_SE=True):

    """
        FUNCTION NAME : read_input_files_hdf5()
        
        DESCRIPTION : 

            Reads the NEMESIS HDF5 input file and fills the parameters in the reference classes.
 
        INPUTS :
      
            runname :: Name of the NEMESIS run

        OPTIONAL INPUTS:
        
            calc_SE :: If True, it will calculate the Measurement error matrix SE (useful only for retrievals, not for forward models)
        
        OUTPUTS : 

            Variables :: Python class defining the parameterisations and state vector
            Measurement :: Python class defining the measurements 
            Atmosphere :: Python class defining the reference atmosphere
            Spectroscopy :: Python class defining the parameters required for the spectroscopic calculations
            Scatter :: Python class defining the parameters required for scattering calculations
            Stellar :: Python class defining the stellar spectrum
            Surface :: Python class defining the surface
            CIA :: Python class defining the Collision-Induced-Absorption cross-sections
            Layer :: Python class defining the layering scheme to be applied in the calculations

        CALLING SEQUENCE:
        
            Atmosphere,Measurement,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval,Telluric = read_input_files_hdf5(runname)
 
        MODIFICATION HISTORY : Juan Alday (25/03/2023)
    """

    from archnemesis import OptimalEstimation_0,Layer_0,Surface_0,Scatter_0,CIA_0,Measurement_0,Spectroscopy_0,Stellar_0, Telluric_0
    import h5py

    #Initialise Atmosphere class and read file
    ##############################################################

    Atmosphere = Atmosphere_0(runname=runname)

    #Read gaseous atmosphere
    Atmosphere.read_hdf5(runname)
    
    #Initialise Layer class and read file
    ###############################################################

    Layer = Layer_0(Atmosphere.RADIUS)
    Layer.read_hdf5(runname)

    #Initialise Surface class and read file
    ###############################################################

    #isurf = planet_info[str(int(Atmosphere.IPLANET))]["isurf"]
    Surface = Surface_0()
    # Always read the Surface entry in HDF5 file as defaults are set
    # there for gas giants that are different for the defaults
    # when constructing via Surface_0()
    Surface.read_hdf5(runname)
    if np.mean(Surface.TSURF)<=0.0:
        Surface.GASGIANT=True   #If T is negative then we omit the surface


    #Initialise Scatter class and read file
    ###############################################################

    Scatter = Scatter_0()
    Scatter.read_hdf5(runname)

    #Initialise CIA class and read files (.cia)  - NOT FROM HDF5 YET
    ##############################################################

    with h5py.File(runname+'.h5','r') as f:
        #Checking if CIA exists
        e = "/CIA" in f
    
    if e==True:
        CIA = CIA_0(runname=runname)
        CIA.read_hdf5(runname)
    else:
        CIA = None

    #Old version of CIA
    #if os.path.exists(runname+'.cia')==True:
    #    CIA = CIA_0(runname=runname)
    #    CIA.read_cia()
    #    #CIA.read_hdf5(runname)
    #else:
    #    CIA = None

    #Initialise Measurement class and read file
    ###############################################################

    Measurement = Measurement_0(runname=runname)
    Measurement.read_hdf5(runname,calc_MeasurementVector=calc_SE)
    
    #Initialise Spectroscopy class and read file
    ###############################################################

    with h5py.File(runname+'.h5','r') as f:
        #Checking if Spectroscopy exists
        e = "/Spectroscopy" in f

    if e is True:
        Spectroscopy = Spectroscopy_0(RUNNAME=runname)
        Spectroscopy.read_hdf5(runname)
    else:
        raise ValueError('error :: Spectroscopy needs to be defined in HDF5 file')

    #Initialise Telluric class and read file
    ###############################################################
    
    with h5py.File(runname+'.h5','r') as f:
        #Checking if Telluric exists
        e = "/Telluric" in f
    
    if e is True:
        Telluric = Telluric_0()
        Telluric.read_hdf5(runname)
    else:
        Telluric = None

    #Reading Stellar class
    ################################################################

    Stellar = Stellar_0()
    Stellar.read_hdf5(runname)

    #Reading .apr file and Variables Class
    #################################################################

    Variables = Variables_0()
    Variables.read_apr(runname, Atmosphere.NP, Atmosphere.NVMR, Atmosphere.NDUST, Atmosphere.NLOCATIONS)
    Variables.XN = copy(Variables.XA)
    Variables.SX = copy(Variables.SA)

    #Reading retrieval setup
    #################################################################

    Retrieval = OptimalEstimation_0()
    Retrieval.read_hdf5(runname)

    return Atmosphere,Measurement,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval,Telluric

###############################################################################################

def read_retparam_hdf5(runname):
    """

        DESCRIPTION : 

            Read the retrieved parameters from the HDF5 file

        INPUTS :
      
            runname :: Name of the NEMESIS run

        OPTIONAL INPUTS: none
        
        OUTPUTS : 

            nvar :: Number of retrieved model parameterisations
            nxvar(nvar) :: Number of parameters associated with each model parameterisation
            varident(nvar,3) :: Variable parameterisation ID
            varparam(nvar,nparam) :: Extra parameters required to model the parameterisations (not retrieved)
            aprparam(nxvar,nvar) :: A priori parameters required to model the parameterisations
            aprerrparam(nxvar,nvar) :: Uncertainty in the a priori parameters required to model the parameterisations 
            retparam(nxvar,nvar) :: Retrieved parameters required to model the parameterisations
            reterrparam(nxvar,nvar) :: Uncertainty in the retrieved parameters required to model the parameterisations


        CALLING SEQUENCE:
        
            nvar,nxvar,varident,varparam,aprparam,aprerrparam,retparam,reterrparam = read_retparam_hdf5(runname)
 
        MODIFICATION HISTORY : Juan Alday (25/03/2023)
    """

    import h5py

    with h5py.File(runname+'.h5','r') as f:

        #Checking if Retrieval exists
        if "/Retrieval" not in f:
            raise ValueError('error :: Retrieval is not defined in HDF5 file')
            return None
        
        if '/Retrieval/Output/Parameters' not in f:
            raise ValueError('error :: Retrieval/Output/Parameters is not defined in HDF5 file')
            return None

        NVAR = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/NVAR', np.int32)
        NXVAR = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/NXVAR', np.array)
        VARIDENT = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/VARIDENT', np.array)
        VARPARAM = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/VARPARAM', np.array)
        RETPARAM = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/RETPARAM', np.array)
        RETERRPARAM = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/RETERRPARAM', np.array)
        APRPARAM = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/APRPARAM', np.array)
        APRERRPARAM = h5py_helper.retrieve_data(f, 'Retrieval/Output/Parameters/APRERRPARAM', np.array)


    return NVAR,NXVAR,VARIDENT,VARPARAM,APRPARAM,APRERRPARAM,RETPARAM,RETERRPARAM

###############################################################################################

def read_bestfit_hdf5(runname):
    """

        DESCRIPTION : 

            Read the best fit from the HDF5 file and include it in the Measurement class

        INPUTS :
      
            runname :: Name of the NEMESIS run

        OPTIONAL INPUTS: none
        
        OUTPUTS : 

            Measurement :: Python class defining the measurement and best fit to the data

        CALLING SEQUENCE:
        
            Measurement = read_bestfit_hdf5(runname)
 
        MODIFICATION HISTORY : Juan Alday (25/03/2023)
    """

    import h5py

    #Reading the best fit
    with h5py.File(runname+'.h5','r') as f:
        if "/Retrieval" not in f:
            raise ValueError('error :: Retrieval is not defined in HDF5 file')
            return None
        
        if '/Retrieval/Output/OptimalEstimation/YN' not in f:
            raise ValueError('error :: Retrieval/Output/OptimalEstimation/YN is not defined in HDF5 file')
            return None
        
        YN = h5py_helper.retrieve_data(f, 'Retrieval/Output/OptimalEstimation/YN', np.array)
 
    #Writing the measurement vector in same format as in Measurement
    Measurement = Measurement_0()
    Measurement.read_hdf5(runname)

    SPECMOD = np.zeros(Measurement.MEAS.shape)
    ix = 0
    for i in range(Measurement.NGEOM):
        SPECMOD[0:Measurement.NCONV[i],i] = YN[ix:ix+Measurement.NCONV[i]]
        ix = ix + Measurement.NCONV[i]

    Measurement.edit_SPECMOD(SPECMOD)

    return Measurement

###############################################################################################
###############################################################################################
#                                     NEMESIS FILES
###############################################################################################
###############################################################################################

# read_input_files()

# read_mre()
# read_cov()
# read_drv()
# read_inp()
# read_set()
# read_fla()
# write_fla()
# write_set()
# write_inp()
# write_err()
# write_fcloud()


###############################################################################################

def read_input_files(runname):

    """
        FUNCTION NAME : read_input_files()
        
        DESCRIPTION : 

            Reads the NEMESIS input files and fills the parameters in the reference classes.
 
        INPUTS :
      
            runname :: Name of the NEMESIS run

        OPTIONAL INPUTS: None
        
        OUTPUTS : 

            Variables :: Python class defining the parameterisations and state vector
            Measurement :: Python class defining the measurements 
            Atmosphere :: Python class defining the reference atmosphere
            Spectroscopy :: Python class defining the parameters required for the spectroscopic calculations
            Scatter :: Python class defining the parameters required for scattering calculations
            Stellar :: Python class defining the stellar spectrum
            Surface :: Python class defining the surface
            CIA :: Python class defining the Collision-Induced-Absorption cross-sections
            Layer :: Python class defining the layering scheme to be applied in the calculations
            Retrieval :: Python class defining the initial Optimal Esimation setup for the input data

        CALLING SEQUENCE:
        
            Atmosphere,Measurement,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval = read_input_files(runname)
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
    """

    #Initialise Atmosphere class and read file (.ref, aerosol.ref)
    ##############################################################

    _lgr.info('Reading atmospheric files...')
    Atm = Atmosphere_0(runname=runname)

    #Read gaseous atmosphere
    Atm.read_ref()

    #Read aerosol profiles
    Atm.read_aerosol()
    
    #Read para-h2
    Atm.read_parah2()
    
    #Read .vpf
    Atm.read_vpf()

    #Reading .set file and starting Scatter, Stellar, Surface and Layer Classes
    #############################################################################

    Layer = Layer_0(Atm.RADIUS)
    Scatter,Stellar,Surface,Layer = read_set(runname,Layer=Layer)
    if Layer.LAYTYP==LayerType.BASE_PRESSURE:
        nlay, pbase = ans.Layer_0.read_play()
        Layer.NLAY = nlay
        Layer.P_base = pbase*101325 
    if Layer.LAYTYP==LayerType.BASE_HEIGHT:
        nlay,hbase = ans.Layer_0.read_hlay()
        Layer.NLAY = nlay
        Layer.H_base = hbase*1.0e3    #Base height of each layer (m)
    if Layer.LAYTYP not in (LayerType.EQUAL_PRESSURE, LayerType.EQUAL_LOG_PRESSURE, LayerType.EQUAL_HEIGHT, LayerType.EQUAL_PATH_LENGTH, LayerType.BASE_PRESSURE, LayerType.BASE_HEIGHT):
        raise ValueError('error in read_input_files :: Need to read the press.lay file but not implemented yet')
    
    Layer.DUST_UNITS_FLAG = Atm.DUST_UNITS_FLAG

    #Reading .inp file and starting Measurement,Scatter and Spectroscopy classes
    #############################################################################

    Measurement,Scatter,Spec,WOFF,fmerrname,NITER,PHILIMIT,NSPEC,IOFF,LIN = read_inp(runname,Scatter=Scatter)

    Measurement.WOFF = WOFF
    
    Retrieval = OptimalEstimation_0()
    Retrieval.NITER=NITER
    Retrieval.PHILIMIT=PHILIMIT
    Retrieval.LIN = LIN

    #Reading surface files if planet has surface
    #############################################################################

    isurf = planet_info[str(int(Atm.IPLANET))]["isurf"]
    if isurf==True:
        if np.mean(Surface.TSURF)>0.0:
            Surface.GASGIANT=False
            Surface.read_sur(runname) #Emissivity (and albedo for Lambert surface)
            if Surface.LOWBC==LowerBoundaryCondition.HAPKE: #Hapke surface
                Surface.read_hap(runname)
        else:
            Surface.GASGIANT=True
    else:
        Surface.GASGIANT=True
    
    if Surface.GASGIANT:
        Surface.LOWBC = LowerBoundaryCondition.THERMAL
        Surface.TSURF = 0.0
        Surface.GALB = 0.0

    #Reading Spectroscopy parameters from .lls or .kls files
    ##############################################################
    if Spec.ILBL==SpectralCalculationMode.K_TABLES:
        Spec.read_kls(runname)
    elif Spec.ILBL==SpectralCalculationMode.LINE_BY_LINE_TABLES:
        Spec.read_lls(runname)
    else:
        raise ValueError('error :: ILBL has to be either SpectralCalculationMode.K_TABLES or SpectralCalculationMode.LINE_BY_LINE_TABLES')

    #Reading extinction and scattering cross sections
    #############################################################################

    Scatter.read_xsc(runname)

    if Scatter.NDUST!=Atm.NDUST:
        raise ValueError('error :: Number of aerosol populations must be the same in .xsc and aerosol.ref files')


    #Initialise Measurement class and read files (.spx, .sha)
    ##############################################################

    Measurement.runname = runname
    Measurement.read_spx()
    
    #Checking if forward modelling error file exists
    if os.path.exists(fmerrname)==True:

        #Reading forward modelling error
        vfmerr,fmerr = read_fmerr(fmerrname)
        Measurement.VFMERR = vfmerr
        Measurement.FMERR = fmerr
        Measurement.add_fmerr()

    #Reading .sha file if FWHM>0.0
    if Measurement.FWHM > 0.0:
        Measurement.read_sha()
    #Reading .fil if FWHM<0.0
    elif Measurement.FWHM < 0.0:
        Measurement.read_fil()

    #Reading stellar spectrum if required by Measurement units
    if Measurement.IFORM in (SpectraUnit.FluxRatio, SpectraUnit.A_Ratio, SpectraUnit.Integrated_spectral_power, SpectraUnit.Atmospheric_transmission):
        Stellar.read_sol(runname)

    #Initialise CIA class and read files (.cia)
    ##############################################################

    if os.path.exists(runname+'.cia')==True:
        CIA = CIA_0(runname=runname)
        CIA.read_cia()
    else:
        CIA = None

    #Reading .fla file
    #############################################################################

    inormal,iray,ih2o,ich4,io3,inh3,iptf,imie,iuv = read_fla(runname)

    if CIA is not None:
        CIA.INORMAL = ParaH2Ratio(inormal)

    Scatter.IRAY = RayleighScatteringMode(iray)
    Scatter.IMIE = AerosolPhaseFunctionCalculationMode(imie)

    if Scatter.ISCAT!=ScatteringCalculationMode.THERMAL_EMISSION:
        if Scatter.IMIE==AerosolPhaseFunctionCalculationMode.HENYEY_GREENSTEIN:
            Scatter.read_hgphase()
        elif Scatter.IMIE==AerosolPhaseFunctionCalculationMode.MIE_THEORY:
            Scatter.read_phase()
        elif Scatter.IMIE==AerosolPhaseFunctionCalculationMode.LEGENDRE_POLYNOMIALS:
            Scatter.read_lpphase()
        else:
            raise ValueError('error :: IMIE must be an integer from 0 to 2')

    #Reading .apr file and Variables Class
    #################################################################
    _lgr.info('Reading .apr file')
    Variables = Variables_0()
    Variables.read_apr(runname, Atm.NP, Atm.NVMR, Atm.NDUST, Atm.NLOCATIONS)
    Variables.XN = copy(Variables.XA)
    Variables.SX = copy(Variables.SA)

    return Atm,Measurement,Spec,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval

###############################################################################################

def read_mre(runname,MakePlot=False):

    """
    FUNCTION NAME : read_mre()

    DESCRIPTION : Reads the .mre file from a Nemesis run

    INPUTS :
    
        runname :: Name of the Nemesis run

    OPTIONAL INPUTS:
    
        MakePlot : If True, a summary plot is made
            
    OUTPUTS : 

        lat :: Latitude (degrees)
        lon :: Longitude (degrees)
        ngeom :: Number of geometries in the observation
        nconv :: Number of points in the measurement vector for each geometry (assuming they all have the same number of points)
        wave(nconv,ngeom) :: Wavelength/wavenumber of each point in the measurement vector
        specret(nconv,ngeom) :: Retrieved spectrum for each of the geometries
        specmeas(nconv,ngeom) :: Measured spectrum for each of the geometries
        specerrmeas(nconv,ngeom) :: Error in the measured spectrum for each of the geometries
        nx :: Number of points in the state vector
        varident(nvar,3) :: Retrieved variable ID, as defined in Nemesis manual
        nxvar :: Number of points in the state vector associated with each retrieved variable
        varparam(nvar,5) :: Extra parameters containing information about how to read the retrieved variables
        aprprof(nx,nvar) :: A priori profile for each variable in the state vector
        aprerr(nx,nvar) :: Error in the a priori profile for each variable in the state vector
        retprof(nx,nvar) :: Retrieved profile for each variable in the state vector
        reterr(nx,nvar) :: Error in the retrieved profile for each variable in the state vector

    CALLING SEQUENCE:

        lat,lon,ngeom,ny,wave,specret,specmeas,specerrmeas,nx,Var,aprprof,aprerr,retprof,reterr = read_mre(runname)
 
    MODIFICATION HISTORY : Juan Alday (15/03/2021)

    """

    #Opening file
    f = open(runname+'.mre','r')

    #Reading first three lines
    tmp = np.fromfile(f,sep=' ',count=1,dtype='int')
    s = f.readline().split()
    #nspec = int(tmp[0])
    tmp = np.fromfile(f,sep=' ',count=5,dtype='float')
    s = f.readline().split()
    #ispec = int(tmp[0])
    ngeom = int(tmp[1])
    ny2 = int(tmp[2])
    ny = int(ny2 / ngeom)
    nx = int(tmp[3])
    tmp = np.fromfile(f,sep=' ',count=2,dtype='float')
    s = f.readline().split()
    lat = float(tmp[0])
    lon = float(tmp[1])
    
    #Reading spectra
    s = f.readline().split()
    s = f.readline().split()
    wave = np.zeros([ny,ngeom])
    specret = np.zeros([ny,ngeom])
    specmeas = np.zeros([ny,ngeom])
    specerrmeas = np.zeros([ny,ngeom])
    for i in range(ngeom):
        for j in range(ny):
            tmp = np.fromfile(f,sep=' ',count=7,dtype='float')
            wave[j,i] = float(tmp[1])
            specret[j,i] = float(tmp[5])
            specmeas[j,i] = float(tmp[2])
            specerrmeas[j,i] = float(tmp[3])

    #Reading the retrieved state vector
    s = f.readline().split()
    if len(s)==2:
        nvar = int(s[1])
    else:
        nvar = int(s[2])

    nxvar = np.zeros([nvar],dtype='int')
    
    Var = Variables_0()
    Var.NVAR = nvar
    aprprof1 = np.zeros([nx,nvar])
    aprerr1 = np.zeros([nx,nvar])
    retprof1 = np.zeros([nx,nvar])
    reterr1 = np.zeros([nx,nvar])
    varident = np.zeros([nvar,3],dtype='int')
    varparam = np.zeros([nvar,5])
    
    for i in range(nvar):
        
        # 1) Read until we find the line that starts with "Variable"
        line = f.readline()
        while line and not line.strip().startswith('Variable'):
            line = f.readline()
        if not line:
            # End of file or no more "Variable" lines
            break

        # 2) The next line is the varident (3 integers)
        line = f.readline().strip()
        parts = line.split()
        if len(parts) != 3:
            raise ValueError(f"Expected 3 integers for varident, got: {parts}")
        varident[i, :] = np.array(parts, dtype=int)

        # 3) The next line is the varparam (5 floats)
        line = f.readline().strip()
        parts = line.split()
        if len(parts) != 5:
            raise ValueError(f"Expected 5 floats for varparam, got: {parts}")
        varparam[i, :] = np.array(parts, dtype=float)
            
        # 4) The next line is typically the header for data lines ("i, ix, xa ...")
        #    We just read it and ignore.
        f.readline()

        # 5) Now read each profile line for this variable
        #    We know the data lines each have 6 columns.
        varlines = []
        while True:
            pos = f.tell()  # Remember position before reading
            line = f.readline()
            if not line:
                break
            if line.strip().startswith('Variable'):
                f.seek(pos)  # Rewind so this line will be read again later
                break
            varlines.append(line)
        nxvar[i] = len(varlines)        

        for j in range(len(varlines)):
            line = varlines[j]
            parts = line.split()
            # If something is off, backtrack and break
            if len(parts) != 6:
                raise ValueError('error in read_mre :: something is off when reading the retrieved parameters')
            aprprof1[j, i] = float(parts[2])
            aprerr1[j, i]  = float(parts[3])
            retprof1[j, i] = float(parts[4])
            reterr1[j, i]  = float(parts[5])

    Var.edit_VARIDENT(varident)
    Var.edit_VARPARAM(varparam)
    Var.NXVAR = nxvar

    aprprof = np.zeros([Var.NXVAR.max(),nvar])
    aprerr = np.zeros([Var.NXVAR.max(),nvar])
    retprof = np.zeros([Var.NXVAR.max(),nvar])
    reterr = np.zeros([Var.NXVAR.max(),nvar])

    for i in range(Var.NVAR):
        aprprof[0:Var.NXVAR[i],i] = aprprof1[0:Var.NXVAR[i],i]
        aprerr[0:Var.NXVAR[i],i] = aprerr1[0:Var.NXVAR[i],i]
        retprof[0:Var.NXVAR[i],i] = retprof1[0:Var.NXVAR[i],i]
        reterr[0:Var.NXVAR[i],i] = reterr1[0:Var.NXVAR[i],i]

    return lat,lon,ngeom,ny,wave,specret,specmeas,specerrmeas,nx,Var,aprprof,aprerr,retprof,reterr

###############################################################################################

def read_cov(runname,MakePlot=False):
    
    
    """
        FUNCTION NAME : read_cov()
        
        DESCRIPTION :
        
            Reads the the .cov file with the standard Nemesis format
        
        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            npro :: Number of points in atmospheric profiles
            nvar :: Number of retrieved variables
            varident(nvar,3) :: Variable ID
            varparam(nvar,mparam) :: Extra parameters for describing the retrieved variable
            nx :: Number of elements in state vector
            ny :: Number of elements in measurement vector
            sa(nx,nx) :: A priori covariance matric
            sm(nx,nx) :: Final measurement covariance matrix
            sn(nx,nx) :: Final smoothing error covariance matrix
            st(nx,nx) :: Final full covariance matrix
            se(ny,ny) :: Measurement error covariance matrix
            aa(nx,nx) :: Averaging kernels
            dd(nx,ny) :: Gain matrix
            kk(ny,nx) :: Jacobian matrix
        
        CALLING SEQUENCE:
        
            npro,nvar,varident,varparam,nx,ny,sa,sm,sn,st,se,aa,dd,kk = read_cov(runname)
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """
    
    #Open file
    f = open(runname+'.cov','r')
    
    #Reading variables that were retrieved
    tmp = np.fromfile(f,sep=' ',count=2,dtype='int')
    npro = int(tmp[0])
    nvar = int(tmp[1])
    
    varident = np.zeros([nvar,3],dtype='int')
    varparam = np.zeros([nvar,5],dtype='int')
    for i in range(nvar):
        tmp = np.fromfile(f,sep=' ',count=3,dtype='int')
        varident[i,:] = tmp[:]
        
        tmp = np.fromfile(f,sep=' ',count=5,dtype='float')
        varparam[i,:] = tmp[:]
    
    
    #Reading optimal estimation matrices
    tmp = np.fromfile(f,sep=' ',count=2,dtype='int')
    nx = int(tmp[0])
    ny = int(tmp[1])


    sa = np.zeros([nx,nx])
    sm = np.zeros([nx,nx])
    sn = np.zeros([nx,nx])
    st = np.zeros([nx,nx])
    aa = np.zeros([nx,nx])
    dd = np.zeros([nx,ny])
    kk = np.zeros([ny,nx])
    se = np.zeros([ny,ny])
    for i in range(nx):
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            sa[i,j] = tmp[0]
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            sm[i,j] = tmp[0]
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            sn[i,j] = tmp[0]
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            st[i,j] = tmp[0]

    for i in range(nx):
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            aa[i,j] = tmp[0]
    
    
    for i in range(nx):
        for j in range(ny):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            dd[i,j] = tmp[0]

    for i in range(ny):
        for j in range(nx):
            tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
            kk[i,j] = tmp[0]
    
    for i in range(ny):
        tmp = np.fromfile(f,sep=' ',count=1,dtype='float')
        se[i,i] = tmp[0]

    f.close()

    return npro,nvar,varident,varparam,nx,ny,sa,sm,sn,st,se,aa,dd,kk

###############################################################################################

def read_drv(runname,MakePlot=False):
    
    """
        FUNCTION NAME : read_drv()
        
        DESCRIPTION : Read the .drv file, which contains all the required information for
                      calculating the observation paths
        
        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS:
        
            MakePlot :: If True, a summary plot is generated
        
        OUTPUTS :
        
            iconv :: Spectral model code
            flagh2p :: Flag for para-H2
            ncont :: Number of aerosol populations
            flagc :: Flag for
            nlayer :: Number of atmospheric layers
            npath :: Number of observation paths
            ngas :: Number of gases in atmosphere
            gasID(ngas) :: RADTRAN gas ID
            isoID(ngas) :: RADTRAN isotopologue ID (0 for all isotopes)
            iproc(ngas) :: Process parameter
            baseH(nlayer) :: Altitude of the base of each layer (km)
            delH(nlayer) :: Altitude covered by the layer (km)
            baseP(nlayer) :: Pressure at the base of each layer (atm)
            baseT(nlayer) :: Temperature at the base of each layer (K)
            totam(nlayer) :: Vertical total column density in atmosphere (cm-2)
            press(nlayer) :: Effective pressure of each layer (atm)
            temp(nlayer) :: Effective temperature of each layer (K)
            doppler(nlayer) ::
            par_coldens(nlayer,ngas) :: Vertical total column density for each gas in atmosphere (cm-2)
            par_press(nlayer,ngas) :: Partial pressure of each gas (atm)
            cont_coldens(nlayer,ncont) :: Aerosol column density for each aerosol population in atmosphere (particles per gram of atm)
            hfp(nlayer) ::
            hfc(nlayer,ncont) ::
            nlayin(npath) :: Number of layers seen in each path
            imod(npath) :: Path model
            errlim(npath) ::
            layinc(npath,2*nlayer) :: Layer indices seen in each path
            emtemp(npath,2*nlayer) :: Emission temperature of each layer in path
            scale(npath,2*nlayer) :: Factor to be applied to the vertical column density to calculate the line-of-sight column density
            nfilt :: Number of profile filter points
            filt(nfilt) :: Filter points
            vfilt(nfilt) ::
            ncalc :: Number of calculations
            itype(ncalc) :: Calculation type
            nintp(ncalc) ::
            nrealp(ncalc) ::
            nchp(ncalc) ::
            icald(ncalc,10) ::
            rcald(ncalc,10) ::
            
        CALLING SEQUENCE:
            
            iconv,flagh2p,ncont,flagc,nlayer,npath,ngas,gasID,isoID,iproc,\
            baseH,delH,baseP,baseT,totam,press,temp,doppler,par_coldens,par_press,cont_coldens,hfp,hfc,\
            nlayin,imod,errlim,layinc,emtemp,scale,\
            nfilt,filt,vfilt,ncalc,itype,nintp,nrealp,nchp,icald,rcald = read_drv(runname)
            
        MODIFICATION HISTORY : Juan Alday (29/09/2019)
            
    """

    f = open(runname+'.drv','r')
    
    #Reading header
    _ = f.readline().split() # header
    var1 = f.readline().split()
    var2 = f.readline().split()
    _ = f.readline().split() # linkey
    
    #Reading flags
    ###############
    flags = f.readline().split()
    iconv = int(flags[0])
    flagh2p = int(flags[1])
    ncont = int(flags[2])
    flagc = int(flags[3])
    
    #Reading name of .xsc file
    _ = f.readline().split() # xscname1
    
    #Reading variables
    ###################
    var1 = f.readline().split()
    nlayer = int(var1[0])
    npath = int(var1[1])
    ngas = int(var1[2])
    
    gasID = np.zeros([ngas],dtype='int32')
    isoID = np.zeros([ngas],dtype='int32')
    iproc = np.zeros([ngas],dtype='int32')
    for i in range(ngas):
        var1 = f.readline().split()
        var2 = f.readline().split()
        gasID[i] = int(var1[0])
        isoID[i] = int(var2[0])
        iproc[i] = int(var2[1])

    #Reading parameters of each layer
    ##################################
    _ = f.readline().split() # header
    _ = f.readline().split() # header
    _ = f.readline().split() # header
    _ = f.readline().split() # header
    baseH = np.zeros([nlayer])
    delH = np.zeros([nlayer])
    baseP = np.zeros([nlayer])
    baseT = np.zeros([nlayer])
    totam = np.zeros([nlayer])
    press = np.zeros([nlayer])
    temp = np.zeros([nlayer])
    doppler = np.zeros([nlayer])
    par_coldens = np.zeros([nlayer,ngas])
    par_press = np.zeros([nlayer,ngas])
    cont_coldens = np.zeros([nlayer,ncont])
    hfp = np.zeros([nlayer])
    hfc = np.zeros([nlayer,ncont])
    for i in range(nlayer):
        #Reading layers
        var1 = f.readline().split()
        baseH[i] = float(var1[1])
        delH[i] = float(var1[2])
        baseP[i] = float(var1[3])
        baseT[i] = float(var1[4])
        totam[i] = float(var1[5])
        press[i] = float(var1[6])
        temp[i] = float(var1[7])
        doppler[i] = float(var1[8])

        #Reading partial pressures and densities of gases in each layer
        nlines = ngas*2./6.
        if nlines-int(nlines)>0.0:
            nlines = int(nlines)+1
        else:
            nlines = int(nlines)

        ix = 0
        var = np.zeros([ngas*2])
        for il in range(nlines):
            var1 = f.readline().split()
            for j in range(len(var1)):
                var[ix] = var1[j]
                ix = ix + 1

        ix = 0
        for il in range(ngas):
            par_coldens[i,il] = var[ix]
            par_press[i,il] = var[ix+1]
            ix = ix + 2
        
        #Reading amount of aerosols in each layer
        nlines = ncont/6.
        if nlines-int(nlines)>0.0:
            nlines = int(nlines)+1
        else:
            nlines = int(nlines)
        var = np.zeros([ncont])
        ix = 0
        for il in range(nlines):
            var1 = f.readline().split()
            for j in range(len(var1)):
                var[ix] = var1[j]
                ix = ix + 1

        ix = 0
        for il in range(ncont):
            cont_coldens[i,il] = var[ix]
            ix = ix + 1

        #Reading if FLAGH2P is set
        if flagh2p==1:
            var1 = f.readline().split()
            hfp[i] = float(var1[0])


        #Reading if FLAGC is set
        if flagc==1:
            var = np.zeros([ncont+1])
            ix = 0
            for il in range(ncont):
                var1 = f.readline().split()
                for j in range(len(var1)):
                    var[ix] = var1[j]
                    ix = ix + 1

            ix = 0
            for il in range(ncont):
                hfc[i,il] = var[ix]
                ix = ix + 1

                    
    #Reading the atmospheric paths
    #########################################
    nlayin = np.zeros([npath],dtype='int32')
    imod = np.zeros([npath])
    errlim = np.zeros([npath])
    layinc = np.zeros([npath,2*nlayer],dtype='int32')
    emtemp = np.zeros([npath,2*nlayer])
    scale = np.zeros([npath,2*nlayer])
    for i in range(npath):
        var1 = f.readline().split()
        nlayin[i] = int(var1[0])
        imod[i] = int(var1[1])
        errlim[i] = float(var1[2])
        for j in range(nlayin[i]):
            var1 = f.readline().split()
            layinc[i,j] = int(var1[1]) - 1   #-1 stands for the fact that arrays in python start in 0, and 1 in fortran
            emtemp[i,j] = float(var1[2])
            scale[i,j] = float(var1[3])

    #Reading number of filter profile points
    #########################################
    var1 = f.readline().split()
    nfilt = int(var1[0])
    filt = np.zeros([nfilt])
    vfilt = np.zeros([nfilt])
    for i in range(nfilt):
        var1 = f.readline().split()
        filt[i] = float(var1[0])
        vfilt[i] = float(var1[1])
                            
    _ = f.readline().split() # outfile

    #Reading number of calculations
    ################################
    var1 = f.readline().split()
    ncalc = int(var1[0])
    itype = np.zeros([ncalc],dtype='int32')
    nintp = np.zeros([ncalc],dtype='int32')
    nrealp = np.zeros([ncalc],dtype='int32')
    nchp = np.zeros([ncalc],dtype='int32')
    icald = np.zeros([ncalc,10],dtype='int32')
    rcald = np.zeros([ncalc,10])
    for i in range(ncalc):
        var1 = f.readline().split()
        itype[i] = int(var1[0])
        nintp[i] = int(var1[1])
        nrealp[i] = int(var1[2])
        nchp[i] = int(var1[3])
        for j in range(nintp[i]):
            var1 = f.readline().split()
            icald[i,j] = int(var1[0])
        for j in range(nrealp[i]):
            var1 = f.readline().split()
            rcald[i,j] = float(var1[0])
        for j in range(nchp[i]):
            var1 = f.readline().split()
            #NOT FINISHED HERE!!!!!!

    f.close()

    if MakePlot==True:
        #Plotting the model for the atmospheric layers
        fig, (ax1,ax2,ax3,ax4) = plt.subplots(1,4,figsize=(15,7))
        ax1.semilogx(baseP,baseH)
        ax1.set_xlabel('Pressure (atm)')
        ax1.set_ylabel('Base altitude (km)')
        ax1.grid()
        
        ax2.plot(baseT,baseH)
        ax2.set_xlabel('Temperature (K)')
        ax2.set_ylabel('Base altitude (km)')
        ax2.grid()
        
        ax3.semilogx(totam,baseH)
        ax3.set_xlabel('Vertical column density in layer (cm$^{-2}$)')
        ax3.set_ylabel('Base altitude (km)')
        ax3.grid()
        
        for i in range(ngas):
            #strgas = spec.read_gasname(gasID[i],isoID[i])
            strgas = 'CHANGE'
            ax4.semilogx(par_coldens[:,i],baseH,label=strgas)
    
        ax4.legend()
        ax4.set_xlabel('Vertical column density in layer (cm$^{-2}$)')
        ax4.set_ylabel('Base altitude (km)')
        ax4.grid()
        
        plt.tight_layout()
        
        plt.show()

    return iconv,flagh2p,ncont,flagc,nlayer,npath,ngas,gasID,isoID,iproc,\
            baseH,delH,baseP,baseT,totam,press,temp,doppler,par_coldens,par_press,cont_coldens,hfp,hfc,\
            nlayin,imod,errlim,layinc,emtemp,scale,\
            nfilt,filt,vfilt,ncalc,itype,nintp,nrealp,nchp,icald,rcald
            
###############################################################################################

def read_inp(runname,Measurement=None,Scatter=None,Spectroscopy=None):

    """
        FUNCTION NAME : read_inp()
        
        DESCRIPTION : Read the .inp file for a Nemesis run

        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            ispace :: (0) Wavenumber in cm-1 (1) Wavelength in um
            iscat :: (0) Thermal emission calculation
                    (1) Multiple scattering required
                    (2) Internal scattered radiation field is calculated first (required for limb-
                        scattering calculations)
                    (3) Single scattering plane-parallel atmosphere calculation
                    (4) Single scattering spherical atmosphere calculation
            ilbl :: (0) Pre-tabulated correlated-k calculation
                    (1) Line by line calculation
                    (2) Pre-tabulated line by line calculation
            
            woff :: Wavenumber/wavelength calibration offset error to be added to the synthetic spectra
            niter :: Number of iterations of the retrieval model required
            philimit :: Percentage convergence limit. If the percentage reduction of the cost function phi
                        is less than philimit then the retrieval is deemed to have converged.
            nspec :: Number of retrievals to perform (for measurements contained in the .spx file)
            ioff :: Index of the first spectrum to fit (in case that nspec > 1).
            lin :: Integer indicating whether the results from previous retrievals are to be used to set any
                    of the atmospheric profiles. (Look Nemesis manual)
        
        CALLING SEQUENCE:
        
            Measurement,Scatter,Spectroscopy,WOFF,fmerrname,NITER,PHILIMIT,NSPEC,IOFF,LIN = read_inp(runname)
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    from archnemesis import Scatter_0, Measurement_0, Spectroscopy_0

    #Getting number of lines 
    nlines = file_lines(runname+'.inp')
    if nlines < 7:
        raise RuntimeError(f"Not enough lines when reading {runname}.inp")
    elif nlines == 7:
        iiform = 0
    elif nlines == 8:
        iiform = 1
    elif nlines > 8:
        iiform = 2

    #Opening file
    f = open(runname+'.inp','r')
    tmp = f.readline().split()
    ispace = WaveUnit(int(tmp[0]))
    iscat = ScatteringCalculationMode(int(tmp[1]))
    ilbl = SpectralCalculationMode(int(tmp[2]))

    if Measurement is None:
        Measurement = Measurement_0()
    Measurement.ISPACE = ispace

    if Scatter is None:
        Scatter = Scatter_0()
    Scatter.ISPACE = ispace
    Scatter.ISCAT = iscat

    if Spectroscopy==None:
        Spectroscopy = Spectroscopy_0(RUNNAME=runname)
    Spectroscopy.ILBL = ilbl

    tmp = f.readline().split()
    WOFF = float(tmp[0])
    fmerrname = str(f.readline().split()[0])
    tmp = f.readline().split()
    NITER = int(tmp[0])
    tmp = f.readline().split()
    PHILIMIT = float(tmp[0])

    tmp = f.readline().split()
    NSPEC = int(tmp[0])
    IOFF = int(tmp[1])

    tmp = f.readline().split()
    LIN = int(tmp[0])

    if iiform == 1:
        tmp = f.readline().split()
        iform = SpectraUnit(int(tmp[0]))
        Measurement.IFORM = iform
    elif iiform == 2:
        tmp = f.readline().split()
        iform = SpectraUnit(int(tmp[0]))
        Measurement.IFORM = iform
        tmp = f.readline().split()
        Measurement.V_DOPPLER = float(tmp[0])
    else:
        Measurement.IFORM = SpectraUnit.Radiance

    return Measurement, Scatter, Spectroscopy, WOFF, fmerrname, NITER, PHILIMIT, NSPEC, IOFF, LIN

###############################################################################################

def read_set(runname,Layer=None,Surface=None,Stellar=None,Scatter=None):
    
    """
        FUNCTION NAME : read_set()
        
        DESCRIPTION : Read the .set file
        
        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            Scatter :: Python class defining the scattering calculations
            Stellar :: Python class defining the stellar properties
            Surface :: Python class defining the surface properties
            Layer :: Python class defining the layering scheme of the atmosphere
        
        CALLING SEQUENCE:
        
            Scatter,Stellar,Surface,Layer = read_set(runname)
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """
    
    from archnemesis import Scatter_0, Stellar_0, Surface_0

    #Opening file
    f = open(runname+'.set','r')
    dummy = f.readline().split()
    nmu1 = f.readline().split()
    nmu = int(nmu1[5])
    mu = np.zeros([nmu],dtype='d')
    wtmu = np.zeros([nmu],dtype='d')
    for i in range(nmu):
        tmp = np.fromfile(f,sep=' ',count=2,dtype='d')
        mu[i] = tmp[0]
        wtmu[i] = tmp[1]
    
    dummy = f.readline().split()
    nf = int(dummy[5])
    dummy = f.readline().split()
    nphi = int(dummy[8])
    dummy = f.readline().split()
    isol = int(dummy[5])
    dummy = f.readline().split()
    dist = float(dummy[5])
    dummy = f.readline().split()
    lowbc = int(dummy[6])
    dummy = f.readline().split()
    galb = float(dummy[3])
    dummy = f.readline().split()
    tsurf = float(dummy[3])

    dummy = f.readline().split()

    dummy = f.readline().split()
    layht = float(dummy[8])
    dummy = f.readline().split()
    nlayer = int(dummy[5])
    dummy = f.readline().split()
    laytp = int(dummy[3])
    dummy = f.readline().split()
    layint = int(dummy[3])

    #Creating or updating Scatter class
    if Scatter==None:
        Scatter = Scatter_0()
        Scatter.NMU = nmu
        Scatter.NF = nf
        Scatter.NPHI = nphi
        Scatter.calc_GAUSS_LOBATTO()
    else:
        Scatter.NMU = nmu
        Scatter.calc_GAUSS_LOBATTO()
        Scatter.NF = nf
        Scatter.NPHI = nphi

    #Creating or updating Stellar class
    if Stellar==None:
        Stellar = Stellar_0()
        Stellar.DIST = dist
        if isol==True:
            Stellar.SOLEXIST = True
            Stellar.read_sol(runname)
        elif isol==False:
            Stellar.SOLEXIST = False
        else:
            raise ValueError('error reading .set file :: SOLEXIST must be either True or False')

    #Creating or updating Surface class
    if Surface==None:
        Surface = Surface_0()

    Surface.LOWBC = LowerBoundaryCondition(lowbc)
    Surface.GALB = galb
    Surface.TSURF = tsurf

    #Creating or updating Layer class
    if Layer==None:
        Layer = Layer_0()
    
    Layer.LAYHT = layht*1.0e3
    Layer.LAYTYP = LayerType(laytp)
    Layer.LAYINT = LayerIntegrationScheme(layint)
    Layer.NLAY = nlayer

    return Scatter,Stellar,Surface,Layer

###############################################################################################

def read_fla(runname):
    
    """
        FUNCTION NAME : read_fla()
        
        DESCRIPTION : Read the .fla file
        
        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            runname :: Name of the Nemesis run
            inormal :: ortho/para-H2 ratio is in equilibrium (0) or normal 3:1 (1)
            iray :: (0) Rayleigh scattering optical depth not included
                    (1) Rayleigh optical depths for gas giant atmosphere
                    (2) Rayleigh optical depth suitable for CO2-dominated atmosphere
                    (>2) Rayleigh optical depth suitable for a N2-O2 atmosphere
            ih2o :: Additional H2O continuum off (0) or on (1)
            ich4 :: Additional CH4 continuum off (0) or on (1)
            io3 :: Additional O3 continuum off (0) or on (1)
            inh3 :: Additional NH3 continuum off (0) or on (1)
            iptf :: Normal partition function calculation (0) or high-temperature partition
                    function for CH4 for Hot Jupiters
            imie :: Only relevant for scattering calculations. (0) Phase function is computed
                    from the associated Henyey-Greenstein hgphase*.dat files. (1) Phase function
                    computed from the Mie-Theory calculated PHASEN.DAT
            iuv :: Additional flag for including UV cross sections off (0) or on (1)
        
        CALLING SEQUENCE:
        
            inormal,iray,ih2o,ich4,io3,inh3,iptf,imie,iuv = read_fla(runname)
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
        """
    
    #Opening file
    f = open(runname+'.fla','r')
    s = f.readline().split()
    inormal = ParaH2Ratio(int(s[0]))
    s = f.readline().split()
    iray = RayleighScatteringMode(int(s[0]))
    s = f.readline().split()
    ih2o = int(s[0])
    s = f.readline().split()
    ich4 = int(s[0])
    s = f.readline().split()
    io3 = int(s[0])
    s = f.readline().split()
    inh3 = int(s[0])
    s = f.readline().split()
    iptf = int(s[0])
    s = f.readline().split()
    imie = AerosolPhaseFunctionCalculationMode(int(s[0]))
    s = f.readline().split()
    iuv = int(s[0])
   
    return inormal,iray,ih2o,ich4,io3,inh3,iptf,imie,iuv

###############################################################################################

def write_fla(runname,inormal,iray,ih2o,ich4,io3,inh3,iptf,imie,iuv):

    """
        FUNCTION NAME : write_fla()
        
        DESCRIPTION : Write the .fla file
        
        INPUTS :
        
            runname :: Name of the Nemesis run
            inormal :: ortho/para-H2 ratio is in equilibrium (0) or normal 3:1 (1)
            iray :: (0) Rayleigh scattering optical depth not included
                    (1) Rayleigh optical depths for gas giant atmosphere
                    (2) Rayleigh optical depth suitable for CO2-dominated atmosphere
                    (>2) Rayleigh optical depth suitable for a N2-O2 atmosphere
            ih2o :: Additional H2O continuum off (0) or on (1)
            ich4 :: Additional CH4 continuum off (0) or on (1)
            io3 :: Additional O3 continuum off (0) or on (1)
            inh3 :: Additional NH3 continuum off (0) or on (1)
            iptf :: Normal partition function calculation (0) or high-temperature partition
                    function for CH4 for Hot Jupiters
            imie :: Only relevant for scattering calculations. (0) Phase function is computed
                    from the associated Henyey-Greenstein hgphase*.dat files. (1) Phase function
                    computed from the Mie-Theory calculated PHASEN.DAT
            iuv :: Additional flag for including UV cross sections off (0) or on (1)
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            Nemesis .fla file       
 
        CALLING SEQUENCE:
        
            write_fla(runname,inormal,iray,ih2o,ich4,io3,inh3,iptf,imie,iuv)
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    f = open(runname+'.fla','w')
    f.write('%i \t %s \n' % (int(inormal),'!INORMAL'))
    f.write('%i \t %s \n' % (int(iray),'!IRAY'))
    f.write('%i \t %s \n' % (ih2o,'!IH2O'))
    f.write('%i \t %s \n' % (ich4,'!ICH4'))
    f.write('%i \t %s \n' % (io3,'!IO3'))
    f.write('%i \t %s \n' % (inh3,'!INH3'))
    f.write('%i \t %s \n' % (iptf,'!IPTF'))
    f.write('%i\t %s \n' % (int(imie),'!IMIE'))
    f.write('%i\t %s \n' % (iuv,'!IUV'))
    f.close()

###############################################################################################

def write_set(runname,nmu,nf,nphi,isol,dist,lowbc,galb,tsurf,layht,nlayer,laytp,layint):

    """
        FUNCTION NAME : write_set()
        
        DESCRIPTION : Read the .set file
        
        INPUTS :
        
            runname :: Name of the Nemesis run
            nmu :: Number of zenith ordinates
            nf :: Required number of Fourier components
            nphi :: Number of azimuth angles
            isol :: Sunlight on/off
            dist :: Solar distance (AU)
            lowbc :: Lower boundary condition (0 Thermal - 1 Lambertian)
            galb :: Ground albedo
            tsurf :: Surface temperature (if planet is not gasgiant)
            layht :: Base height of lowest layer
            nlayer :: Number of vertical levels to split the atmosphere into
            laytp :: Flag to indicate how layering is perfomed (radtran)
            layint :: Flag to indicate how layer amounts are calculated (radtran)

        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            Nemesis .set file

        CALLING SEQUENCE:
        
            l = write_set(runname,nmu,nf,nphi,isol,dist,lowbc,galb,tsurf,layht,nlayer,laytp,layint)
        
        MODIFICATION HISTORY : Juan Alday (15/10/2019)
        
    """

    #Calculating the Gauss-Lobatto quadtrature points
    iScatter = Scatter_0(NMU=nmu)

    #Writin the .set file
    f = open(runname+'.set','w')
    f.write('********************************************************* \n')
    f.write('Number of zenith angles : '+str(nmu)+' \n')
    for i in range(nmu):
        f.write('\t %10.12f \t %10.12f \n' % (iScatter.MU[i],iScatter.WTMU[i]))
    f.write('Number of Fourier components : '+str(nf)+' \n')
    f.write('Number of azimuth angles for fourier analysis : '+str(nphi)+' \n')
    f.write('Sunlight on(1) or off(0) : '+str(isol)+' \n')
    f.write('Distance from Sun (AU) : '+str(dist)+' \n')
    f.write('Lower boundary cond. Thermal(0) Lambert(1) : '+str(lowbc)+' \n')
    f.write('Ground albedo : '+str(galb)+' \n')
    f.write('Surface temperature : '+str(tsurf)+' \n')
    f.write('********************************************************* \n')
    f.write('Alt. at base of bot.layer (not limb) : '+str(layht)+' \n')
    f.write('Number of atm layers : '+str(nlayer)+' \n')
    f.write('Layer type : '+str(laytp)+' \n')
    f.write('Layer integration : '+str(layint)+' \n')
    f.write('********************************************************* \n')

    f.close()

###############################################################################################

def write_inp(runname,ispace,iscat,ilbl,woff,niter,philimit,nspec,ioff,lin,IFORM=-1):

    """
        FUNCTION NAME : write_inp()
        
        DESCRIPTION : Write the .inp file for a Nemesis run
        
        INPUTS :
        
            runname :: Name of the Nemesis run
            ispace :: (0) Wavenumber in cm-1 (1) Wavelength in um
            iscat :: (0) Thermal emission calculation
                    (1) Multiple scattering required
                    (2) Internal scattered radiation field is calculated first (required for limb-
                        scattering calculations)
                    (3) Single scattering plane-parallel atmosphere calculation
                    (4) Single scattering spherical atmosphere calculation
            ilbl :: (0) Pre-tabulated correlated-k calculation
                    (1) Line by line calculation
                    (2) Pre-tabulated line by line calculation
            woff :: Wavenumber/wavelength calibration offset error to be added to the synthetic spectra
            niter :: Number of iterations of the retrieval model required
            philimit :: Percentage convergence limit. If the percentage reduction of the cost function phi
                        is less than philimit then the retrieval is deemed to have converged.
            nspec :: Number of retrievals to perform (for measurements contained in the .spx file)
            ioff :: Index of the first spectrum to fit (in case that nspec > 1).
            lin :: Integer indicating whether the results from previous retrievals are to be used to set any
                    of the atmospheric profiles. (Look Nemesis manual)
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
        CALLING SEQUENCE:

            write_inp(runname,ispace,iscat,ilbl,woff,niter,philimit,nspec,ioff,lin,IFORM=iform)
         
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    #Opening file
    f = open(runname+'.inp','w')
    f.write('%i \t %i \t %i \n' % (int(ispace), int(iscat), int(ilbl)))
    f.write('%10.5f \n' % (woff))
    f.write(runname+'.err \n')
    f.write('%i \n' % (niter))
    f.write('%10.5f \n' % (philimit))
    f.write('%i \t %i \n' % (nspec, ioff))
    f.write('%i \n' % (lin))
    if IFORM != -1:
        f.write('%i \n' % (int(IFORM)))
    f.close()


###############################################################################################

def read_pre(runname):
    
    
    """
        FUNCTION NAME : read_pre()

        DESCRIPTION :
        
            Reads the .pre file for a Nemesis run.
            
            This file includes information about the retrieved state vector and the retrieved
            covariance matrix. This file is generated as an output file in the retrieval. 
            If it is copied to a .pre extension and LIN>0 in the .inp file, this file will be used
            to implement information from previous retrievals into the new retrieval
            
            NOTE :: The implementation of this file in archNEMESIS is slightly different
                    than in NEMESIS, but the purpose of the file remains the same.
        
        INPUTS :
        
            runname :: Name of the Nemesis run
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            Variables_prev :: Variables class with the information from the previous retrieval
            
        
        CALLING SEQUENCE:

            Variables_prev = read_pre(runname)

        MODIFICATION HISTORY : Juan Alday (29/09/2025)
        
    """
    
    #Reading all lines first
    with open(runname + '.pre', 'r') as f: 
        # Skip all lines that do not start with '#' 
        lines = [line.strip() for line in f.readlines() if line.strip()]
    
    #Reading file
    idx = 0
    nspec = int(lines[idx].split()[0])
    _lgr.debug(f'nspec = {nspec}')
    idx += 1
    if nspec != 1:
        raise ValueError('error in read_raw :: nspec>1 and this has not yet been implemented in archnemesis')
    
    for ispec in range(nspec):
    
        #Reading lines
        ispecx = int(lines[idx].split()[0])
        idx += 1
        
        #Reading latitude and longitude
        lat = float(lines[idx].split()[0])
        lon = float(lines[idx].split()[1])
        idx += 1
        _lgr.debug(f'lat = {lat}, lon = {lon}')
        
        #Reading other parameters
        npro = int(lines[idx].split()[0])
        ngas = int(lines[idx].split()[1])
        ndust = int(lines[idx].split()[2])
        nlocations = int(lines[idx].split()[3])
        nvar = int(lines[idx].split()[4])
        idx += 1
        _lgr.debug(f'npro = {npro}, ngas = {ngas}, ndust = {ndust}, nlocations = {nlocations}, nvar = {nvar}')
        
        #Reading variable ids
        varident = np.zeros((nvar,3),dtype='int32')
        varparam = np.zeros((nvar,5))
        for ivar in range(nvar):
            
            idx += 1  # Skipping header line
            
            for j in range(3):
                varident[ivar,j] = int(lines[idx].split()[j])
            idx += 1
            
            for j in range(5):
                varparam[ivar,j] = float(lines[idx].split()[j])
            idx += 1
            
            _lgr.debug(f'varident[{ivar}] = {varident[ivar,:]}, varparam[{ivar}] = {varparam[ivar,:]}')
            
        #Reading state vector
        nx = int(lines[idx].split()[0])
        idx += 1
        _lgr.debug(f'nx = {nx}')
        
        xn = np.zeros(nx)
        lx = np.zeros(nx,dtype='int32')
        num = np.zeros(nx,dtype='int32')
        for ix in range(nx):
            xn[ix] = float(lines[idx].split()[0])
            lx[ix] = int(lines[idx].split()[1])
            num[ix] = int(lines[idx].split()[2])
            idx += 1
            
        #Reading the retrieved coviariance matrix
        st = np.zeros((nx,nx))
        for ix in range(nx):
            for iy in range(nx):
                st[ix,iy] = float(lines[idx].split()[0])
                idx += 1
            
    #Creating variables class 
    Variables_prev = ans.Variables_0()    
    Variables_prev._models = []
    jsurf = -1
    jalb = -1
    jxsc = -1
    jtan = -1
    jpre = -1
    #jrad = -1
    jlogg = -1
    jfrac = -1
    ix = 0
    for ivar in range(nvar):
        
        for model in Models:
            
            if model.is_varident_valid(varident[ivar,:]):
                found_model_for_varident = True
                
                try:
                    Variables_prev._models.append(
                        model.from_bookmark(
                            Variables_prev, 
                            varident[ivar,:], 
                            varparam[ivar,:], 
                            ix, 
                            npro, 
                            ngas,
                            ndust,
                            nlocations,
                        )
                    )
                except Exception as e:
                    raise ValueError(f'Failed to read {ivar}^th model entry (with VARIDENT={varident[ivar,:]})') from e
                
                _lgr.info(f'\nVariables_0 :: read_raw :: varident {varident[ivar,:]}. Constructed model "{model.__name__}" (id={model.id})')
                
                if varident[ivar][2]==999:  #Retrieval of surface temperature
                    jsurf = ix
                elif varident[ivar][2]==666: #Retrieval of pressure at a givent tangent height
                    jpre = ix
                elif varident[ivar][2]==777: #Retrieval of tangent height at a given pressure level
                    jtan = ix
                
        ix += Variables_prev._models[-1].n_state_vector_entries

        if not found_model_for_varident:
            raise ValueError(f'Variables_0 :: read_apr :: no model found for varident {varident[ivar,:]}')

        
    Variables_prev.NVAR=nvar
    Variables_prev.NPARAM=5
    Variables_prev.edit_VARIDENT(varident)
    Variables_prev.edit_VARPARAM(varparam)
    Variables_prev.calc_NXVAR(npro,nlocations=nlocations)
    Variables_prev.JPRE, Variables_prev.JTAN, Variables_prev.JSURF, Variables_prev.JALB, Variables_prev.JXSC, Variables_prev.JLOGG, Variables_prev.JFRAC = jpre, jtan, jsurf, jalb, jxsc, jlogg, jfrac
    Variables_prev.NX = nx
    Variables_prev.edit_XA(xn)
    Variables_prev.edit_XN(xn)
    Variables_prev.edit_SA(st)
    Variables_prev.edit_LX(lx)
    Variables_prev.NUM = num
    Variables_prev.calc_DSTEP()
    Variables_prev.calc_FIX()
    
    return Variables_prev


###############################################################################################

def read_fmerr(filename):
    
    
    """
        FUNCTION NAME : read_fmerr()

        DESCRIPTION :
        
            Reads the forward modelling error from a file
        
        INPUTS :
        
            filename :: Name of the file containing the forward modelling error
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            verr(nwave) :: Wavenumber/Wavelength array
            fwerr(nwave) :: Forward modelling error
        
        CALLING SEQUENCE:

            verr,fwerr = read_fmerr(filename)

        MODIFICATION HISTORY : Juan Alday (29/09/2025)
        
    """
    
    with open(filename, 'r') as f:
        
        for line in f:
            if line.strip() and not line.startswith('#'):
                nwave = int(line.split()[0])
                break
            
        verr = np.zeros(nwave)
        fwerr = np.zeros(nwave)
        i = 0
        for line in f:
            if line.strip() and not line.startswith('#'):
                values = line.split()
                if len(values) >= 2:
                    verr[i] = float(values[0])
                    fwerr[i] = float(values[1])
                    i += 1
                if i >= nwave:
                    break

    return verr, fwerr

###############################################################################################

def write_err(runname,nwave,wave,fwerr):

    """
        FUNCTION NAME : write_err()
        
        DESCRIPTION : Write the .err file, including information about forward modelling error
        
        INPUTS :
        
            runname :: Name of Nemesis run
            nwave :: Number of wavelengths at which the albedo is defined
            wave(nwave) :: Wavenumber/Wavelength array
            fwerr(nwave) :: Forward modelling error
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            Nemeis .err file       
 
        CALLING SEQUENCE:
        
            write_err(runname,nwave,wave,fwerr)
         
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    f = open(runname+'.err','w')
    f.write('\t %i \n' % (nwave))
    for i in range(nwave):
        f.write('\t %10.5f \t %10.5f \n' % (wave[i],fwerr[i]))
    f.close()

###############################################################################################

def write_fcloud(npro,naero,height,frac,icloud, MakePlot=False):
    
    """
        FUNCTION NAME : write_fcloud()
        
        DESCRIPTION : Writes the fcloud.ref file, which specifies if the cloud is in the form of
                      a uniform thin haze or is instead arranged in thicker clouds covering a certain
                      fraction of the mean area.
        
        INPUTS :
        
            npro :: Number of altitude profiles in reference atmosphere
            naero :: Number of aerosol populations in the atmosphere
            height(npro) :: Altitude (km)
            frac(npro) :: Fractional cloud cover at each level
            icloud(npro,naero) :: Flag indicating which aerosol types contribute to the broken cloud
                                  which has a fractional cloud cover of frac
        
        OPTIONAL INPUTS: None
        
        OUTPUTS :
        
            fcloud.ref file
        
        CALLING SEQUENCE:
        
            write_fcloud(npro,naero,height,frac,icloud)
        
        MODIFICATION HISTORY : Juan Alday (16/03/2021)
        
    """

    f = open('fcloud.ref','w')

    f.write('%i \t %i \n' % (npro,naero))
    
    for i in range(npro):
        str1 = str('{0:7.6f}'.format(height[i]))+'\t'+str('{0:7.3f}'.format(frac[i]))
        for j in range(naero):
            str1 = str1+'\t'+str('{0:d}'.format(icloud[i,j]))
            f.write(str1+'\n')

    f.close()







from __future__ import annotations #  for 3.9 compatability

from typing import Self

import numpy as np

import archnemesis.helpers.maths_helper as maths_helper
from archnemesis.database.datatypes.gas_descriptor import RadtranGasDescriptor


class GasMixture:
    """
    Class for storing a gas mixture and getting properties of each gas within the mixture.
    
    The fractions of a gas in the mixture always add to one, they are normalised by the
    `normalise()` method, and after every `set(...)` method.
    """
    def __init__(self, gas_frac_pairs : tuple[tuple[RadtranGasDescriptor, float],...]):
        self._gf = dict(gas_frac_pairs)
        self.normalise()
    
    @property
    def gasses(self) -> tuple[RadtranGasDescriptor]:
        return tuple(self._gf.keys())
    
    @property
    def fractions(self) -> tuple[float]:
        return tuple(self._gf.values())
    
    def __contains__(self, k) -> bool:
        return k in self._gf
    
    def __getitem__(self, k) -> float:
        return self._gf[k]
    
    def normalise(self):
        if len(self._gf) == 0:
            return
        
        a = np.sum([f for g,f in self._gf.items()])
        self._gf = dict([(g,f/a) for g,f in self._gf.items()])
    
    def is_normalised(self) -> bool: # ideally this should always be True when using the public interface of this class
        if len(self._gf) == 0:
            return True
        
        a = np.sum([f for g,f in self._gf.items()])
        return (a - 1) < 1E-30
    
    def set(self, gas_frac_pairs : tuple[tuple[RadtranGasDescriptor, float],...]) -> Self:
        for g, f in gas_frac_pairs:
            self._gf[g] = f
        
        self.normalise()
        return self
    
    def get_number_densities(self, press : float | np.ndarray, temp : float | np.ndarray) -> dict[RadtranGasDescriptor, float | np.ndarray]:
        n = maths_helper.ideal_gas_number_density(press, temp)
        
        ns = dict()
        for g, f in self._gf.items():
            ns[g] = n * f
        return ns
    
    
    


from __future__ import annotations #  for 3.9 compatability

from typing import Iterator

from . import PreRTModels
from . import PostRTModels
from .ModelBase import ModelBase

import inspect


import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)


def _get_all_model_classes():
    """
    Pulls all model classes from their files
    """
    return (
        *tuple(
            x[1] for x in inspect.getmembers(
                PreRTModels,
                lambda x: inspect.isclass(x) and issubclass(x, ModelBase) and x.id is not None
            )
        ),
        *tuple(
            x[1] for x in inspect.getmembers(
                PostRTModels,
                lambda x: inspect.isclass(x) and issubclass(x, ModelBase) and x.id is not None
            )
        ),
    )


class _ModelPrinterMixin(type):
    def __repr__(cls) -> str:
        return '\n\n'.join((x.to_string() for x in cls._model_classes.values()))
    
    def __str__(cls) -> str:
        return '\n\n'.join((x.to_string() for x in cls._model_classes.values()))
    
    def __iter__(cls) -> Iterator[ModelBase]:
        yield from cls._model_classes.values()
    
    def __getitem__(cls, id) -> ModelBase:
        return cls._model_classes[id]
    
    def __contains__(cls, id) -> bool:
        return id in cls._model_classes
    
    @property
    def ids(cls) -> tuple[int,...]:
        return tuple(x.id for x in cls._model_classes.values())
    
    @property
    def names(cls) -> tuple[str,...]:
        return tuple(x.__name__ for x in cls._model_classes.values())
    
    def info(cls, id=None) -> str:
        if id is None:
            return repr(cls)
        
        return cls._model_classes[id].to_string()


class Models(metaclass = _ModelPrinterMixin):
    """
    Holds all of the model parameterisation classes indexed by ID number
    
    ## PROPERTIES ##
    
        ids -> tuple[int,...]
            Returns a tuple of all the ids for all the model classes stored by this class
        
        names -> tuple[str,...]
            Returns a tuple of the class names of all the model classes stored by this class
            
    ## METHODS ##
        
        info([id : int]) -> str
            Returns a string with a summary of information for each model class stored by this class,
            if passed an `id` will only return information string for the model class with that id.
        
        __str__() -> str
        __repr__() -> str
            Returns the same information as `info()`, but enables you to write `_lgr.info(Models)`
            as syntactic sugar
        
        __iter__() -> Iterator[ModelBase]
            Returns an iterator that loops over each model class stored by this class. 
            Example:
                ```
                for model_class in Models:
                    _lgr.info(model_class.parameters)
                ```
        
        __getitem__(id) -> ModelBase
            Returns the model class with `id` stored by this model class, or raises `KeyError` if
            no class with `id` exists.
        
        __contains__(id) -> bool
            Returns True if a model class with `id` is stored, else returns False.
        
        display(id) -> None
            Prints the `info()` entry for stored model class with `id` to stdout.
        
        
        
    """
    
    _model_classes = dict((x.id,x) for x in _get_all_model_classes())
    
    @classmethod
    def display(cls, id) -> None:
        _lgr.info(cls[id].to_string())
    
    @classmethod
    def as_string(cls, id) -> str:
        return cls[id].to_string()
    




from __future__ import annotations #  for 3.9 compatability

import os
import os.path
import pickle
import warnings

import numpy as np
import numpy.ma


import archnemesis as ans
import archnemesis.enums
import archnemesis.database.wrappers.hapi as hapi
from ..protocols import (
    LineDatabaseProtocol, 
    LineDataProtocol, 
)
from ..datatypes.wave_range import WaveRange
#from ..datatypes.gas_isotopes import GasIsotopes
from ..datatypes.gas_descriptor import RadtranGasDescriptor

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.INFO)

# NOTE: HAPI does not differentiate between an actual failure to retrieve data because of a problem vs
#       not retrieving any data because there is no data in a wavelength range. Until I can handle those
#       cases differently, there will always be some annoying exceptions when no lines are present in a 
#       requested wavelength range.

class HITRAN(LineDatabaseProtocol):
    """
    Class that implements the "LineProviderProtocol" for HITRAN datasets.
    
    NOTE: In the background are using one local table per gas, therefore we can use 'local_iso_id' to uniquely determine
    which isotopologue we are talking about when working from local tables. If this changes, we need to re-think how this
    is done.
    
    NOTE: HAPI does not do databases properly so we cannot get a handle to the actual database connection, we need to 
    ensure that `hapi.db_begin(...)` is only called once so we need class variables to make sure the class can remember
    if it has initialised the database or not, and for instances to pass information between eachother. In the background
    there is **only ever one database** so if we want to move to a different database while running we have to somehow
    "forget" the old one, and the re-initialise on the new one (including the time taken to parse the large column text
    files that act as the "database" on-disk storage).
    
    NOTE: There is not way to get HAPI to update a table it will always overwrite it, therefore we want to have a single
    table for each isotope.
    
    TODO: HAPI does not store the last request, therefore we cannot know how many wavelengths were requested. We therefore
    keep track of the limits of the wave range requested for each (gas, ambient-gas) pair. We always request the 
    wave interval that encompasses the cached interval and the requested wave range. That means if you ask for 
    (100,200) cm^{-1} and the cached range is (9000, 10000) cm^{-1}, then the re-requested range will be the encompassing
    interval (100, 10000) cm^{-1}. We do this becase HAPI always overwrites the previously downloaded data, it cannot
    update it.
    """
    
    
    _class_local_storage_dir : str = os.path.abspath('local_line_database')
    
    _class_downloaded_gas_wavenumber_interval_cache_file : str = 'downloaded_gas_wavenumber_interval.pkl'
    
    _class_db_init_flag : bool = False
    
    _class_downloaded_gas_wavenumber_interval : dict[tuple[RadtranGasDescriptor, ans.enums.AmbientGas], WaveRange] = dict() 
        
    _class_gas_wavenumber_interval_to_download : dict[tuple[RadtranGasDescriptor, ans.enums.AmbientGas], WaveRange] = dict() 
    
    
    @classmethod
    def set_class_local_storage_dir(cls, local_storage_dir : str):
        if cls._class_db_init_flag:
            raise RuntimeError('Cannot change location of HITRAN database after it has been initialised as HAPI does not have a way of using multiple databases at once.')
        else:
            cls._class_local_storage_dir = os.path.abspath(local_storage_dir)
    
    
    @classmethod
    def set_class_db_init_flag(cls, v : bool):
        cls._class_db_init_flag = v
    
    
    @property
    def local_storage_dir(self) -> str:
        """
        Required by LineDatabaseProtocol. Gets the directory the local database is in.
        """
        return self._class_local_storage_dir
    
    @local_storage_dir.setter
    def local_storage_dir(self, value : str) -> None:
        """
        Required by LineDatabaseProtocol. Sets the directory the local database is in.
        """
        self.set_class_local_storage_dir(value)
    
    @property
    def db_init_flag(self) -> bool:
        """
        Gets the value of the flag that tell us if the local database has been initialised
        """
        return self._class_db_init_flag
    
    @db_init_flag.setter
    def db_init_flag(self, value : bool) -> None:
        """
        Sets the value of the flag that tell us if the local database has been initialised
        """
        self.set_class_db_init_flag(value)
    
    @property
    def _downloaded_gas_wavenumber_interval(self) -> dict[RadtranGasDescriptor, WaveRange]:
        return self._class_downloaded_gas_wavenumber_interval
    
    @property
    def _gas_wavenumber_interval_to_download(self) -> dict[RadtranGasDescriptor, WaveRange]:
        return self._class_gas_wavenumber_interval_to_download
    
    @property
    def _downloaded_gas_wavenumber_interval_cache_file(self):
        return os.path.join(self._class_local_storage_dir, self._class_downloaded_gas_wavenumber_interval_cache_file)
    
    
    @staticmethod
    def get_tablename_from_props(gas_desc : RadtranGasDescriptor, ambient_gas : ans.enums.AmbientGas) -> str:
        return f'{gas_desc.gas_name}_{gas_desc.iso_id}_ambient_{ambient_gas.name}'
    
    @staticmethod
    def get_props_from_tablename(tablename : str) -> tuple[RadtranGasDescriptor, ans.enums.AmbientGas]:
        gas_name, iso_id, _, amb_gas = tablename.split('_')
        return (
            RadtranGasDescriptor(
                getattr(ans.enums.Gas, gas_name),
                int(iso_id)
            ),
            getattr(ans.enums.AmbientGas, amb_gas),
        )
    
    @staticmethod
    def get_ambient_gas_parameter_name_strings(ambient_gas : ans.enums.AmbientGas) -> tuple[str,str,str]:
        """
        HITRAN paramter names have gas in UPPERCASE (except for 'air')
        """
        if ambient_gas == ans.enums.AmbientGas.AIR:
            gamma_str = 'gamma_air'
            n_str = 'n_air'
            delta_str = 'delta_air'
        elif ambient_gas == ans.enums.AmbientGas.CO2:
            gamma_str = 'gamma_CO2'
            n_str = 'n_CO2'
            delta_str = 'delta_CO2'
        else:
            gamma_str = f'gamma_{ambient_gas.name.upper()}'
            n_str = f'n_{ambient_gas.name.upper()}'
            delta_str = f'delta_{ambient_gas.name.upper()}'
            #raise ValueError(f'Unrecognised ambient gas {ambient_gas}')
        
        return gamma_str, n_str, delta_str
    
    @staticmethod
    def get_ambient_gas_column_name_strings(ambient_gas : ans.enums.AmbientGas) -> tuple[str,str,str]:
        """
        HAPI column names have gas in LOWERCASE (including 'air')
        """
        if ambient_gas == ans.enums.AmbientGas.AIR:
            gamma_str = 'gamma_air'
            n_str = 'n_air'
            delta_str = 'delta_air'
        elif ambient_gas == ans.enums.AmbientGas.CO2:
            gamma_str = 'gamma_co2'
            n_str = 'n_co2'
            delta_str = 'delta_co2'
        else:
            gamma_str = f'gamma_{ambient_gas.name.lower()}'
            n_str = f'n_{ambient_gas.name.lower()}'
            delta_str = f'delta_{ambient_gas.name.lower()}'
            #raise ValueError(f'Unrecognised ambient gas {ambient_gas}')
        
        return gamma_str, n_str, delta_str
    
    def __init__(
            self,
            local_storage_dir : None | str = None,
        ):
        self.init_database(local_storage_dir)
    
    @property
    def ready(self) -> bool:
        """
        Required by LineDatabaseProtocol. Returns True if the database is ready to use, False otherwise
        """
        return self.db_init_flag
    
    def purge(self) -> None:
        """
        Required by LineDatabaseProtocol. Remove all local data and make it so the database must be reinitalised
        """
        if not self.db_init_flag:
            _lgr.warning('HITRAN database is not initialised yet, so cannot purge it.')
            return
        
        tablenames = list(hapi.tableList())
        
        
        # drop all tables
        for tablename in tablenames:
            hapi.dropTable(tablename)
        hapi.db_commit()
        
        # Remove any table files remaining
        for tablename in tablenames:
            table_header_path = os.path.join(self.local_storage_dir, f'{tablename}.header')
            table_data_path = os.path.join(self.local_storage_dir, f'{tablename}.data')
            table_fpaths = (
                table_header_path, 
                table_data_path
            )
            for fpath in table_fpaths:
                if os.path.exists(fpath):
                    os.remove(fpath)
                
        # delete all cache files
        cache_fpaths = (
            self._downloaded_gas_wavenumber_interval_cache_file,
        )
        for fpath in cache_fpaths:
            if os.path.exists(fpath):
                os.remove(fpath)
        
        # Delete local storage of data
        for k in tuple(self._downloaded_gas_wavenumber_interval.keys()):
            del self._downloaded_gas_wavenumber_interval[k]
        
        for k in tuple(self._gas_wavenumber_interval_to_download.keys()):
            del self._gas_wavenumber_interval_to_download[k]
        
        # Need this hack so tables are not held on to by `hapi`
        hapi.LOCAL_TABLE_CACHE = dict()
        
        # set flag so that database can be reinitalised
        self.db_init_flag = False
        _lgr.info(f'Database {self} has been purged. It no longer holds any local data and must be re-initialised using `self.init_database(...)`.')
        
        return


    def get_line_data(
            self, 
            gas_descs : tuple[RadtranGasDescriptor,...], 
            wave_range : WaveRange, 
            ambient_gas : ans.enums.AmbientGas
        ) -> dict[RadtranGasDescriptor, LineDataProtocol]:
        """
        Required by LineDatabaseProtocol. Retrieves line data from the database.
        
        Checks if database has been initialised, checks if we have all requested data, download if required, reads
        requested data from database and returns it in LineDataProtocol format
        """
        gd = tuple(gas_descs)
        if not self.db_init_flag:
            self._init_database()
        
        self._check_available_data(gd, wave_range, ambient_gas)
        self._fetch_line_data()
        
        return self._read_line_data(
            gd, 
            wave_range, 
            ambient_gas
        )
    
    def retrieve_downloaded_gas_wavenumber_interval_from_cache(self):
        cache_file_path = self._downloaded_gas_wavenumber_interval_cache_file
        
        if os.path.exists(cache_file_path):
            _lgr.debug(f'Loading `self._downloaded_gas_wavenumber_interval` from {cache_file_path=}')
            loaded = dict()
            with open(cache_file_path, 'rb') as f:
                try:
                    loaded = pickle.load(f)
                except Exception as e:
                    _lgr.warning(f'Something went wrong when unpickling `self._downloaded_gas_wavenumber_interval` from "{cache_file_path}", assuming no cached data. Error: {str(e)}')
                else:
                    self._downloaded_gas_wavenumber_interval.update(loaded)
            _lgr.debug(f'Loaded cached {self._downloaded_gas_wavenumber_interval=}')
        else:
            _lgr.info(f'Cache file for `self._downloaded_gas_wavenumber_interval` not found at {cache_file_path=}. This is not a problem if the database is being built from scratch.')

    def store_downloaded_gas_wavenumber_interval_to_cache(self):
        cache_file_path = self._downloaded_gas_wavenumber_interval_cache_file
        
        _lgr.debug(f'Caching {self._downloaded_gas_wavenumber_interval=} to {cache_file_path=}')
        with open(cache_file_path, 'wb') as f:
            try:
                pickle.dump(self._downloaded_gas_wavenumber_interval, f)
            except Exception as e:
                _lgr.warning(f'Something went wrong when pickling `self._downloaded_gas_wavenumber_interval` to "{cache_file_path}". Data will not be cached. Error: {str(e)}')


    def init_database(
            self,
            local_storage_dir : None | str = None,
        ):
        
        
        if self.db_init_flag:
            _lgr.warning(f'Database {self} is already initialised, cannot initialise it again unless it is purged using `self.purge()`.')
            return
        
        if local_storage_dir is not None:
            self.local_storage_dir = local_storage_dir
        
        # Create directory to store database in
        os.makedirs(self.local_storage_dir, exist_ok=True)
        
        # Read downloaded_gas_wavenumber_interval_cache_file if it exists
        self.retrieve_downloaded_gas_wavenumber_interval_from_cache()
        

        #Starting the HAPI database
        hapi.db_begin(self.local_storage_dir)
        tablenames = list(hapi.tableList())
        
        _lgr.debug(f'{tablenames=}')
        
        # Read all of the data we have stored locally
        for tablename in tablenames:
            if _lgr.level <= logging.DEBUG:
                hapi.describeTable(tablename)
            
            # Assume tables conform to our naming format
            try:
                gas_desc, ambient_gas = self.get_props_from_tablename(tablename)
            except Exception:
                _lgr.warning(f'HITRAN database found table "{tablename}" that does not conform to our naming format. Skipping...')
                continue
            
            # Should only ever have a single gas isotope per file
            hapi.select(tablename, ParameterNames=('molec_id', 'local_iso_id', 'nu'), Conditions=None, DestinationTableName = 'temp')
            found_molec_ids, found_iso_local_ids, found_v = hapi.getColumns('temp', ('molec_id', 'local_iso_id', 'nu'))
            
            assert len(set(found_molec_ids)) == 1, "Should only have one HITRAN gas per table"
            assert len(set(found_iso_local_ids)) == 1, "Should only have one HITRAN isotope per table"
            
            
            found_v = np.array(found_v, dtype=float)
            
            vmin = np.min(found_v)
            vmax = np.max(found_v)
            
            gda_pair = (gas_desc, ambient_gas)
            if gda_pair not in self._downloaded_gas_wavenumber_interval:
                self._downloaded_gas_wavenumber_interval[gda_pair] = WaveRange(vmin, vmax, ans.enums.WaveUnit.Wavenumber_cm)
                _lgr.info(f'No cache of downloaded wave range for {gda_pair}, using wave range {self._downloaded_gas_wavenumber_interval[gda_pair]=} found in tables {tablename} which will underestimate the requested range.')
                
            hapi.dropTable('temp')
        
        self.db_init_flag = True
        _lgr.info(f'Database {self} initialised.')
        
        return

    def _check_available_data(self, gas_descs : tuple[RadtranGasDescriptor,...], wave_range : WaveRange, ambient_gas : ans.enums.AmbientGas):
        for gas_desc in gas_descs:
            gda_pair = (gas_desc, ambient_gas)
            if gda_pair not in self._downloaded_gas_wavenumber_interval:
                self._gas_wavenumber_interval_to_download[gda_pair] = wave_range.to_unit(ans.enums.WaveUnit.Wavenumber_cm)
            elif not self._downloaded_gas_wavenumber_interval[gda_pair].contains(wave_range):
                self._gas_wavenumber_interval_to_download[gda_pair] = self._downloaded_gas_wavenumber_interval[gda_pair].union(wave_range)
    
    
    def _fetch_line_data(self) -> dict[bool,...]:
        if not self.db_init_flag:
            raise RuntimeError(f'Cannot fetch line data as database {self} is not initalised yet.')
        
        gas_exists_in_db = [False]*len(self._gas_wavenumber_interval_to_download)
        
        for gas_idx, (gda_pair, wave_range) in enumerate(tuple(self._gas_wavenumber_interval_to_download.items())):
            
            # Check that the gas exists in the HITRAN database
            gas_desc, ambient_gas = gda_pair
            ht_gas = gas_desc.to_hitran()
            if ht_gas is None:
                gas_exists_in_db[gas_idx] = False
                _lgr.warning(f'Cannot download data for {gas_desc}, as that gas is not present in the HITRAN database')
                continue # cannot download data for a gas that is not present in HITRAN database
            else:
                gas_exists_in_db[gas_idx] = True
            
            # Work out if we have to download anything, or if we can just use the data we already have
            if gda_pair not in self._gas_wavenumber_interval_to_download:
                continue
            saved_wave_range = self._downloaded_gas_wavenumber_interval.get(gda_pair, None)
            if saved_wave_range is not None and saved_wave_range.contains(wave_range):
                _lgr.debug(f'Downloaded gas data {gda_pair} CONTAINS desired wave range {saved_wave_range} vs {wave_range}')
                continue
            else:
                _lgr.debug(f'Downloaded gas data {gda_pair} DOES NOT CONTAIN desired wave range {saved_wave_range} vs {wave_range}')
        
            
            # If we must download data then perform the download here.
            
            _lgr.info(f'Downloading data for {gda_pair} where {wave_range=} ({saved_wave_range=})...')
            
            vmin, vmax = wave_range.as_unit(ans.enums.WaveUnit.Wavenumber_cm).values()
            
            gamma_str, n_str, delta_str = self.get_ambient_gas_parameter_name_strings(ambient_gas)
            parameters = (
                'nu',
                'sw',
                'a',
                gamma_str,
                n_str,
                delta_str,
                'gamma_self',
                'n_self', # NOTE: This may be missing, in that case it is assumed to be equal to n_air
                'elower',
                'n_air', # NOTE: here so we always fetch this value as we do not know if 'n_self' will be different or not
            )
            _lgr.debug(f'Fetching the following parameters from HITRAN: {parameters}')
            try:
                hapi.fetch(
                    self.get_tablename_from_props(gas_desc, ambient_gas),
                    ht_gas.gas_id,
                    ht_gas.iso_id,
                    vmin,
                    vmax,
                    Parameters=parameters
                )
            except Exception as e:
                msg = '\n'.join((
                    'HAPI failed to retrieve data from HITRAN servers, possible reasons are listed below:',
                    '',
                    f'    1) There may be no transitions (lines) in your chosen wavelength range, try again with a different range. {wave_range=}',
                    '',
                    f'    2) One of the parameter names may be misspelled (unlikely), the parameter names used were: {parameters}',
                    '',
                    f'    3) There may be no gas with the requested gas_id {ht_gas.gas_id} and iso_id {ht_gas.iso_id} numbers, see "https://hitran.org/docs/iso-meta/" for accepted hitran gas codes (NOTE: these are different to Radtran gas codes, but they should be converted internally)',
                    '',
                    '    4) The internet connection may have dropped',
                    '',
                    'For more detailed troubleshooting change line 22 of archnemesis/database/wrappers/hapi.py to "_lgr.setLevel(logging.DEBUG)" and re-run. More detailed output from HAPI will be displayed.'
                ))
                raise RuntimeError(msg) from e
            else:
                hapi.db_commit()
            
            _lgr.info('Data downloaded.')
            
            self._downloaded_gas_wavenumber_interval[gda_pair] = wave_range
            if gda_pair in self._gas_wavenumber_interval_to_download:
                del self._gas_wavenumber_interval_to_download[gda_pair]
            
        # Cache to downloaded_gas_wavenumber_interval_cache_file for future
        self.store_downloaded_gas_wavenumber_interval_to_cache()
        
        return gas_exists_in_db


    


    def _read_line_data(self, gas_descs : tuple[RadtranGasDescriptor,...], wave_range : WaveRange, ambient_gas : ans.enums.AmbientGas):
        if not self.db_init_flag:
            raise RuntimeError(f'Cannot read line data as database {self} is not initalised yet.')
        
        temp_line_data_table_name = 'temp_line_data'
        _lgr.debug(f'{temp_line_data_table_name=}')
        
        line_data = dict()
        
        _lgr.debug(f'{gas_descs=}')
        
        for gas_desc in gas_descs:
            _lgr.debug(f'{gas_desc=}')
            ht_gas_desc = gas_desc.to_hitran()
            if ht_gas_desc is None:
                line_data[gas_desc] = None
                continue
            
            if _lgr.level <= logging.DEBUG:
                hapi.describeTable(self.get_tablename_from_props(gas_desc, ambient_gas))
            
            vmin, vmax = wave_range.values()
            Conditions = ('and',('between', 'nu', vmin, vmax),('equal','local_iso_id',ht_gas_desc.iso_id))
            
            try:
                hapi.select(
                    self.get_tablename_from_props(gas_desc, ambient_gas), 
                    Conditions=Conditions, 
                    DestinationTableName=temp_line_data_table_name
                )
            except Exception as e:
                raise RuntimeError(f'Failure when reading from database {self}.') from e
        
            if _lgr.level <= logging.DEBUG:
                hapi.describeTable(temp_line_data_table_name)
        
            gamma_str, n_str, delta_str = self.get_ambient_gas_column_name_strings(ambient_gas)
            col_names = (
                'nu',
                'sw',
                'a',
                gamma_str,
                n_str,
                delta_str,
                'gamma_self',
                'n_self', # NOTE: This may be missing, in that case it is assumed to be equal to n_air
                'elower',
                'n_air', # NOTE: here as we always fetch this value as we do not know if 'n_self' will be different or not
            )
            
            cols = hapi.getColumns(
                temp_line_data_table_name,
                col_names
            )
            
            with warnings.catch_warnings():
                warnings.simplefilter('default', UserWarning)
                
                # If we don't have any 'n_self' values, replace 'n_self' with final 'n_air' column
                # Then, drop final 'n_air' column as we don't need it
                n_self_tmp = np.ma.array(cols[-3], dtype=float)
                n_self = np.array(cols[-1], dtype=float)
                n_self[~n_self_tmp.mask] = n_self_tmp[~n_self_tmp.mask]
                
                cols[-3] = n_self
                cols = cols[:-1]
                
                _lgr.debug(f'{len(cols)=} {[len(c) for c in cols]=}')
                
                for i, c in enumerate(cols):
                    missing_col_names = []
                    if isinstance(c[0], np.ma.core.MaskedConstant):
                        missing_col_names.append(col_names[i])
                    if len(missing_col_names) > 0:
                        raise RuntimeError(f'Gas {gas_desc.gas_name} isotope {gas_desc.iso_id} "{gas_desc.isotope_name}" with ambient gas "{ambient_gas.name}" has NULL entries in the HITRAN database for the following columns: {missing_col_names}')
            
                line_data[gas_desc] = np.array(
                    list(zip(*cols)),
                    dtype = [
                        ('NU', float), # Transition wavenumber (cm^{-1})
                        ('SW', float), # transition intensity (weighted by isotopologue abundance) (cm^{-1} / molec_cm^{-2})
                        ('A', float), # einstein-A coeifficient (s^{-1})
                        ('GAMMA_AMB', float), # ambient gas broadening coefficient (cm^{-1} atm^{-1})
                        ('N_AMB', float), # temperature dependent exponent for `gamma_amb` (NUMBER)
                        ('DELTA_AMB', float), # ambient gas pressure induced line-shift (cm^{-1} atm^{-1})
                        ('GAMMA_SELF', float), # self broadening coefficient (cm^{-1} atm^{-1})
                        ('N_SELF', float), # temperature dependent exponent for `gamma_self` (NUMBER)
                        ('ELOWER', float), # lower state energy (cm^{-1})
                    ]
                ).view(np.recarray)
            
            hapi.dropTable(temp_line_data_table_name)
        
        return line_data
    
    
    def _read_partition_function_data(self, gas_descs : tuple[RadtranGasDescriptor,...]):
        partition_function_data = dict()
        
        for gas_desc in gas_descs:
            ht_gas = gas_desc.to_hitran()
            if ht_gas is None:
                partition_function_data[gas_desc] = None
                continue
            
            temps = hapi.TIPS_2021_ISOT_HASH[(ht_gas.gas_id,ht_gas.iso_id)]
            qs = hapi.TIPS_2021_ISOQ_HASH[(ht_gas.gas_id,ht_gas.iso_id)]
            partition_function_data[gas_desc] = np.array(
                list(zip(
                    temps,
                    qs
                )),
                dtype=[
                    ('TEMP', float), 
                    ('Q', float)
                ]
            ).view(np.recarray)
        return partition_function_data
    



import pytest  
import archnemesis as ans
import numpy as np

from archnemesis.Models import _get_all_model_classes

def test_model_classes_have_unique_id_numbers():
    """
    All model classes should have a unique ID number.
    """
    seen_id_numbers = dict()
    for model_class in _get_all_model_classes():
        seen_id_numbers[model_class.id] = (*seen_id_numbers.get(model_class.id, tuple()), model_class)
    
    for id, model_classes in seen_id_numbers.items():
        assert len(model_classes) == 1,f"Model id {id} is not unique, is used by all of these models: {tuple(m.__module__+'.'+m.__qualname__ for m in model_classes)}"


def test_model_classes_have_no_abstract_methods():
    """
    The aim is to make sure model classes can be instantiated, and that we didn't miss any abstract methods when
    writing a model class.
    """
    
    model_instances = []
    
    for model_class in _get_all_model_classes():
        assert len(model_class.__abstractmethods__) == 0, f'Model id {model_class.id} must not have any abstract methods left (must be a concrete class). Has abstract methods: {tuple(model_class.__abstractmethods__)}'





import warnings
import pytest
import archnemesis as ans
import importlib
from pathlib import Path
import os

def test_no_syntax_warnings_on_compile(recwarn):
    caught_warnings = []
    
    for path in ans.__path__:
        for dirpath, dirnames, filenames in os.walk(str(path)):
            for filename in filenames:
                if filename.endswith('.py'):
                    file = Path(dirpath) / filename
                    loader = importlib.machinery.SourceFileLoader("<test_import>", file)
                    loader.source_to_code(loader.get_data(file), file)

    if len(recwarn) > 0:
        msg = "\n".join(map(lambda w: f'{w.category.__name__} {w.filename} {w.lineno}: {w.message}', filter(lambda w: issubclass(w.category, SyntaxWarning), recwarn)))
        assert len(msg) == 0, f"Caught an warnings on compile, test failed:\n{msg}"
